#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Objects defined in the Scripting API specification represented as classes that are basically dict-wrappers.

.. autosummary::
    :toctree: _dictionaries

    wotpy.wot.dictionaries.base
    wotpy.wot.dictionaries.filter
    wotpy.wot.dictionaries.interaction
    wotpy.wot.dictionaries.link
    wotpy.wot.dictionaries.schema
    wotpy.wot.dictionaries.security
    wotpy.wot.dictionaries.thing
    wotpy.wot.dictionaries.version
"""
