#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
WebSockets Protocol Binding implementation.

.. autosummary::
    :toctree: _ws

    wotpy.protocols.ws.client
    wotpy.protocols.ws.enums
    wotpy.protocols.ws.handler
    wotpy.protocols.ws.messages
    wotpy.protocols.ws.schemas
    wotpy.protocols.ws.server
"""
