#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
HTTP Protocol Binding implementation.

.. autosummary::
    :toctree: _http

    wotpy.protocols.http.handlers
    wotpy.protocols.http.client
    wotpy.protocols.http.enums
    wotpy.protocols.http.server
"""
