#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
HTTP request handlers that implement each of the Interaction verbs.

.. autosummary::
    :toctree: _handlers

    wotpy.protocols.http.handlers.action
    wotpy.protocols.http.handlers.event
    wotpy.protocols.http.handlers.property
    wotpy.protocols.http.handlers.utils
"""
